"""Dummy methods if devices not attached."""


def t_in(_board_num, _channel, _scale, _options=0):
    """Get dummy value for `t_in`."""
    return 1.0


def v_in(_board_num, _channel, _ul_range, _options=0):
    """Get dummy value for `v_in`."""
    return 1.0
